/*
* Sergey Reva (rs_rlab@mail.ru)
* http://rs-rlab.narod.ru/
*/

#include <u.h>
#include <libc.h>
#include <thread.h>
#include <auth.h>

#define TIME asctime(localtime(time(nil)))

enum {
	MaxNode=64,
};

typedef struct {
	char *name;
	int perf;
} Node;

char *charset, *key, *target, *port="33";
Node nodes[MaxNode];
int nodecount=0, maxkeylen=-1, keylen, setlen;
Channel *chankey;

void usage(void)
{
	print("usage %s: -v -t file -l maxkeylen -c charset -s start key -p port node1 node2 ... nodeN\n",argv0);
	print("\t* in verbose mode (v) every second you get statistic\n");
	print("\t* host should have running service 'slave -s' at port 'port', default is 33\n");
	threadexitsall(0);
}

int openfile(Node *n, char *file, int omode)
{
	char name[128];

	snprint(name, 128, "/n/%s/%s", n->name, file);
	return open(name, omode);
}

int mounthost(Node* n)
{
	int fd ,i;
	char path[128];

	for(i=0; i<5; i++, sleep(2000))
	{
		fd=dial(netmkaddr(n->name, 0, port), 0, 0, 0);

		if (fd<0)
			continue;

		snprint(path,128,"/n/%s",n->name);
		return amount(fd, path, MAFTER, "");	
	}

	return -1;
}

void unmounthost(Node *n)
{
	char path[128];

	snprint(path,128,"/n/%s",n->name);
	unmount(nil, path);
}

int upload(int fd, char* src)
{
	int srcfd, size;
	char buf[1024];

	srcfd=open(src, OREAD);
	if (srcfd<0)
		return -1;

	while(size=read(srcfd, buf, 1024))
		write(fd, buf, size);

	close(srcfd);
	return 1;
}

void slaveproc(void *p)
{
	Node *n=(Node*)p;
	char *testkey, buf[128];
	int fd, size;

	n->perf=0;

	print("Connecting to %s\n",n->name);
	if (mounthost(n)<0)
	{
		print("Host %s offline\n", n->name);
		threadexits(0);
	}

	print("Host %s online\n", n->name);

	fd=openfile(n, "data", OWRITE);
	if (fd<0)
	{
		print("Host %s down (data)...\n", n->name);
		unmounthost(n);
		threadexits(0);
	}

	if (upload(fd, target)<0)
	{
		print("Host %s can't recieve data\n", n->name);
		close(fd);
		unmounthost(n);
		threadexits(0);
	}
	close(fd);

	testkey=malloc(maxkeylen);

	while(recv(chankey, testkey)==1)
	{
		n->perf++;

		fd=openfile(n, "key", OWRITE);
		if (fd<0)
		{
			print("Host %s down (key)...\n", n->name);			
			free(testkey);
			unmounthost(n);
			threadexits(0);
		}
		write(fd, testkey, strlen(testkey));
		close(fd);

		fd=openfile(n, "ctl", OREAD);
		if (fd<0)
		{
			print("Host %s down (ctl)...\n", n->name);
			free(testkey);
			unmounthost(n);
			threadexits(0);
		}
		size=read(fd, buf, 128);
		close(fd);
		buf[size]=0;
		
		if (!strcmp(buf, "valid"))
			print("Host %s found key %s at %s", n->name, testkey, TIME);	
	}

	free(testkey);
	threadexits(0);
}

void sortcharset(void)
{
	int i, j;
	char t;
	setlen=strlen(charset);

	for(j=1;j;)
	for (j=0,i=0;i<setlen-1;i++)
		if(charset[i]>charset[i+1])
		{
			t=charset[i+1];
			charset[i+1]=charset[i];
			charset[i]=t;
			j++;
		}
}

char nextsym(char c)
{
	int i;
	
	if (charset[setlen-1]==c)
		return 0;

	for(i=0;i<setlen-1;i++)
		if(charset[i]==c)
			return charset[i+1];

	return charset[0];
}

void generatekey(void)
{
	int i;

	for(;keylen<=maxkeylen;)
	{
		key[keylen]=0;
		send(chankey, key);

		for(i=keylen-1;i>=0;i--)
		{
			key[i]=nextsym(key[i]);
			if (key[i])
				break;

			key[i]=charset[0];

			if (i==0)
			{
				keylen++;
				if (keylen<=maxkeylen)
				{
					memmove(key+1,key,keylen-1);
					key[keylen-1]=charset[0];
				}
				break;	
			}
		}
	}
}

void statproc(void*)
{
	char *testkey, buf[4096];
	int n, i;

	testkey=malloc(maxkeylen);

	while(recv(chankey, testkey)==1)
	{
		send(chankey, testkey);

		n=0;
		n+=sprint(buf+n, "*** Last key: %s ", testkey);

		for(i=0; i<nodecount; i++)
			n+=sprint(buf+n, "%s: %d ", nodes[i].name, nodes[i].perf);

		print("%s***\n", buf);

		sleep(1000);
	}

	free(testkey);
	threadexits(0);
}

void threadmain(int argc, char *argv[])
{
	int i, verbose=0;
	char *initkey=nil, t[2];

	ARGBEGIN {
	case 'p':
		port=ARGF();
		break;
	case 'v':
		verbose++;
		break;
	case 't':
		target=ARGF();
		break;
	case 'l':
		maxkeylen=atoi(ARGF());
		break;
	case 'c':
		charset=ARGF();
		break;
	case 's':
		initkey=ARGF();
		break;
	} ARGEND

	while((*argv) && (nodecount<MaxNode))
	{
		nodes[nodecount].name=*argv;
		nodecount++;
		argv++;
	}

	if (!charset)
		usage();

	sortcharset();

	if (maxkeylen<=0)
		usage();
	
	key=malloc(maxkeylen+1);
	if (initkey!=nil)
	{
		keylen=strlen(initkey);
		memmove(key, initkey, keylen);
	}
	else
	{
		key[0]=charset[0];
		keylen=1;
	}
	key[keylen]=0;

	chankey=chancreate(maxkeylen+1, 128);
	if (!chankey)
	{
		free(key);
		threadexitsall("can't create channel");
	}

	print("Charset: %s\nMax length: %d\nStart from: %s\n%d hosts requested\n",
		charset, maxkeylen, key, nodecount);

	print("Start at %s", TIME);

	if (verbose)
		proccreate(statproc, nil, mainstacksize);

	for(i=0;i<nodecount;i++)
		proccreate(slaveproc, &nodes[i], mainstacksize);

	generatekey();

	print("All done at %s", TIME);

	chanfree(chankey);
	threadexits(0);
}
